# Attendance ID generator
# Copyright 2024-01-16 Osamu NAKAMURA

# RUN
#    e.g.> python 01_id_generator.py 15 3
#             For 3 classes of 15 students.
#    e.g.> python 01_id_generator.py
#             If the arguments are omitted, the default number of
#             students and classes is used.

# ADJUST
#  - n_std : Number of IDs to be generated per lecture (= Total number of enrollments).
#  - n_lec : Number of lectures for IDs to be generated.
#  - Adjust the format of attedance ID prints to fit your printer. See below.

# OUTPUT
#  - id_001.csv         /id_002.csv         /id_003.csv          ... :
#            Files that store attendance IDs. 001/002/.. are lec-ID.
#  - id_001_printer.xlsx/id_002_printer.xlsx/id_003_printer.xlsx/... :
#            Files for printing.

# Printing paper information:
# Various perforated papers are available at ASKUL.
# https://www.askul.co.jp/p/3998878/?int_id=recom_DtVar

import random
import sys
import csv
import pandas    # Installation is required.
import openpyxl  # Installation is required.



#################################################
def rndid(id_ltr):
        # function to generate a random ID.
        # The argument is a list containing the characters used for the ID.

    n_rnd1 = random.randrange(0, len(id_ltr), 1)
    n_rnd2 = random.randrange(0, len(id_ltr), 1)
    n_rnd3 = random.randrange(0, len(id_ltr), 1)
    n_rnd4 = random.randrange(0, len(id_ltr), 1)
    n_rnd5 = random.randrange(0, len(id_ltr), 1)
    n_rnd6 = random.randrange(0, len(id_ltr), 1)
    n_rnd7 = random.randrange(0, len(id_ltr), 1)
    n_rnd8 = random.randrange(0, len(id_ltr), 1)
        # Randomly generate integers between 0 and id_ltr-1.

    tmp = id_ltr[n_rnd1] + id_ltr[n_rnd2] + "-" \
        + id_ltr[n_rnd3] + id_ltr[n_rnd4] + "-" \
        + id_ltr[n_rnd5] + id_ltr[n_rnd6] + "-" \
        + id_ltr[n_rnd7] + id_ltr[n_rnd8]
        # Generate ID.
    
    return tmp


#################################################
def convert_1d_to_2d(dt, cols):
        # Convert a one-D list into a two-D list.

        # This function is taken from the following site:
        # https://note.nkmk.me/python-list-ndarray-1d-to-2d/
    
    return [dt[i:i + cols] for i in range(0, len(dt), cols)]


#################################################

############
############ Parameter Setting
############

n_std  = 300
        # The default Number of IDs to be generated per lecture (= Total number of enrollments).
        # This variable is overwritten if arguments are given below.
        # .._std: student
n_lec  =  14
        # The default Number of lectures for IDs to be generated.
        # This variable is overwritten if arguments are given below.
n_col  =   3
        # Number of columns in the output Excel file.

n_rnd  =  10
        # Maximum number of iterations when the same ID is generated.
        # If this number is exceeded, the program terminates as an error.

id_ltr = [ "2", "3", "4", "5", "6", "7", "8", "9",      \
           "a", "c", "d", "e", "f", "g", "k", "m", "p", "s", "t", "w", "x", "y"  ]
        # Characters to be used for the generated ID.
        # Avoid characters that are difficult to distinguish from each other, such as "1" and "l".

#id_ltr = [ "2", "3" ]     # For test.
#id_ltr = [ "2"      ]     # For test.

cl_exl = [ "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L" ]
        # Excel column IDs, simply...

id_std = [""]  * ( n_std * n_lec )
        # Initialize the list that will store the IDs to be generated.

f_bas  = "./id_"
        # Base name of output files.
        # "./" represents the current directory.


v_arg  = sys.argv
        # read arguments.

if ( len(v_arg) >= 3 ):
    n_std = int( v_arg[1] )
    n_lec = int( v_arg[2] )
        # If 2 arguments (excluding this prog.) are given,
        # they are assigned to n_std and n_lec.


############
############ Check the probability of coincidence.
############ Check n_col setting.
############

n_cov  = float( len(id_ltr)**8 ) / float(n_std * n_lec)
        # Calculate how many times the total number of IDs available is the total number of enrollments.
        # If this number is smaller than about ~10000, redundancy is poor and the possibility of coincidence increases.

print("\n"+"Total number of IDs that can be assigned per enrollee: ", n_cov, "\n")


if ( n_col > len(cl_exl) ):
    print("\n" + "Error: The number set in n_col must not exceed the number of elements in cl_exl." + "\n")
    sys.exit()


############
############ Generate random IDs
############

for i_std in range(n_std * n_lec):
        # Loop for each enrollment for all lecture sessions.
    
    for i_rnd in range(n_rnd):
        # Loop to create a random ID.
        # Usually assumed to be completed once.
        # Loop if the same ID has been created previously.

        id_std[i_std] = rndid(id_ltr)
        # Generate a random ID.

        #print(i_std, id_std[i_std])
        # For debug.

        ##
        ## Check to see if the same ID has been created previously.
        ##

        flg_chk = 0
        for i_tmp in range(i_std):
            #print(i_tmp, i_std)
            if ( id_std[i_tmp] == id_std[i_std] ):
                print(" Same ID found.")
                flg_chk = 1
                break
            # If the same ID is found, set flg_chk to 1 and break the loop.

        if ( flg_chk == 0 ):
            break
            # If flg_chk=0, i.e., the same ID is not found, break the
            # i_rnd-loop and finish random ID generation.
            # If not, continue the loop and regenerate the ID.

    if ( i_rnd == n_rnd-1 and flg_chk == 1 ):
        print("\n" + "Error: All different IDs could not be created within the specified number of n_rnd.\n")
        sys.exit()
        # Check to see if the i_rnd-loop has run to the end with flg_chk=1 (flg_chk=1 means the same ID exists).
        # If it goes all the way to the end, it has failed to eliminate the same ID within n_rnd times trial.
        # In that case, an error is output.
        # I expect that this will not normally be the case.


############
############ Output to Excel
############
        
for i_lec in range(n_lec):
        # Loop for each lecture.

    id_nw1  = id_std[ i_lec * n_std : ( i_lec + 1 ) * n_std ]
        # Cut out the IDs for the i_lec-th lecture from id_std.
        # Note that dt[1:3] cuts out dt[1]--dt[2] while dt[3] is not included.
        
    #print(i_lec, len(id_now), id_now[0], id_now[len(id_now)-1] )
        # For debug.

    id_nw2  = [ "Class-" + "{:0>3}".format(i_lec+1) + "\n" + s for s in id_nw1 ]
        # Add the lecture times information at the beginning of the ID.
        # - Since i_lec starts counting from 0, add 1.
    
    id_nw3  = convert_1d_to_2d( id_nw2, n_col )
        # Convert id_nw2, a one-dimensional list, into a two-dimensional list.
    
    id_nw1d = pandas.DataFrame(id_nw1)
    id_nw3d = pandas.DataFrame(id_nw3)
        # Convert 2-D list into pandas dataframe.

    f_ot1   = f_bas + "{:0>3}".format(i_lec+1) +         ".csv"
    f_ot2   = f_bas + "{:0>3}".format(i_lec+1) + "_printer.xlsx"
        # Output file name.
        # - Since i_lec starts counting from 0, add 1.
        # - e.g. {:0>3}.format(15) --> 015

    id_nw1d.to_csv(   f_ot1,                       index=False, header=False )
    id_nw3d.to_excel( f_ot2, sheet_name="printer", index=False, header=False )
        # Output to csv/excel file.
        # - index =False to not write out line   names.
        # - header=False to not write out column names.
        # - f_ot1 = simple IDs to csv
        # - f_ot2 = IDs in a printable format to xlsx.


    ##
    ## Formatting (Excel file)
    ##

    book   = openpyxl.load_workbook(f_ot2) # Obtain the book.
    sheet  = book[ "printer" ]             # Obtain the sheet.

    for i_c in range( n_col ):
        sheet.column_dimensions[cl_exl[i_c]    ].width  = 29 # Edit to fit your printer and paper.
        # Specify column width.

    for i_r in range( len(id_nw3) ):
        sheet.row_dimensions[          i_r + 1 ].height = 74 # Edit to fit your printer and paper.
        # Specify the height of the row.
    
    for i_c in range( n_col ):
     for i_r in range( len(id_nw3) ):
        sheet[ cl_exl[i_c] + str(i_r+1) ].alignment = \
            openpyxl.styles.Alignment( horizontal="center", vertical="center", wrapText=True )
        # - Specify horizontal/vertical alignment.
        # - wrapText=T: Fold back to show the whole.
        # - cl_exl[i_c]+str(i_r+1) = "A1", "A2", .., "B1", "B2", ..
        sheet[ cl_exl[i_c] + str(i_r+1) ].font      = \
            openpyxl.styles.Font( name="Verdana", size=12, bold=True) 
        # Specify font.
    
    sheet.page_margins.top                 = 0.84        # number in inch unit.
    sheet.page_margins.bottom              = 0.84        # number in inch unit.
    sheet.page_margins.right               = 0.626       # number in inch unit.
    sheet.page_margins.left                = 0.626       # number in inch unit.
    sheet.page_margins.header              = 0.0         # number in inch unit.
    sheet.page_margins.footer              = 0.0         # number in inch unit.
    sheet.page_setup.orientation           = "portrait"
    sheet.print_options.horizontalCentered = True
    sheet.print_options.verticalCentered   = True
        # printer settings.

    book.save(f_ot2)
        # Save the formatting.
